class CVS(object):

    def __init__(self):
        super().__init__()

    def GetLocalDir(self):
        raise NotImplementedError()

    def GetCurrentRevision(self, branch=None):
        raise NotImplementedError()

    def GetRevisionFromTag(self, tag):
        raise NotImplementedError()

    def GetCurrentBranch(self):
        raise NotImplementedError()

    def GetFirstBranchCommit(self, branch=None):
        raise NotImplementedError()

    def GetCommitSinceBranch(self, branch=None, master="master"):
        raise NotImplementedError()

    def GetPatchPathsFromCommit(self, commit):
        raise NotImplementedError()

    def GetCherryPickedCommit(self, commit, start, branch=None):
        raise NotImplementedError()

    def GetAllBranchCommits(self, branch=None):
        raise NotImplementedError()

    def GetUpstreamChangeSet(self, start=None, end=None, branch=None):
        raise NotImplementedError()



from ensurecvs.arg_parser import parse_args
from ensurecvs.classes.CVS import CVSHelper
from ensurecvs.classes.IssueFinder import IssueFinder

if __name__ == '__main__':
    runargs = parse_args()
    __cvs = CVSHelper.GetCVS(runargs.localdir)

    _startrev = runargs.srcrev or __cvs.GetRevisionFromTag(
        runargs.srctag) or None

    _masterSinceBranch = __cvs.GetCommitSinceBranch(
        branch=runargs.srcbranch, master=runargs.upstream)
    _masterSinceBranch = [x for x in _masterSinceBranch if not __cvs.GetCherryPickedCommit(
        x, _masterSinceBranch[0].hexsha, branch=runargs.srcbranch)]
    _upstreamChanges = __cvs.GetUpstreamChangeSet(
        branch=runargs.srcbranch, start=_startrev)
    _upstreamChanges = [x for x in _upstreamChanges if not __cvs.GetCherryPickedCommit(
        x, _upstreamChanges[0].hexsha, branch=runargs.upstream)]

    _findingsUpstream = IssueFinder(useDefaultFilter=runargs.usedefaultfiler,
                                    filterLists=runargs.filterLists).GetIssues(__cvs, _upstreamChanges)
    _findingsMaster = IssueFinder(useDefaultFilter=runargs.usedefaultfiler,
                                  filterLists=runargs.filterLists).GetIssues(__cvs, _masterSinceBranch)
    for f in _findingsUpstream:
        print("[{}] commit {}:'{}' is likely to contain bugfixes".format(
            runargs.srcbranch or __cvs.GetCurrentBranch(), f.hexsha, f.summary))

    for f in _findingsMaster:
        print("[{}] commit {}:'{}' is likely to contain bugfixes".format(
            runargs.upstream, f.hexsha, f.summary))



import argparse


def parse_args():
    global RUNARGS
    global DEFAULT_VERSION
    parser = argparse.ArgumentParser(
        prog="ensurecsv", description="Ensure that you're using the most security source code")
    parser.add_argument("localdir", help="Path to local repo")
    parser.add_argument("--srcbranch", default=None,
                        help="Use explicitly given branch")

    revgroup = parser.add_mutually_exclusive_group()
    revgroup.add_argument("--srcrev", default=None,
                          help="Use explicitly given source revision")
    revgroup.add_argument("--srctag", default=None,
                          help="Use explicitly given tag")

    parser.add_argument("--upstream", default="master",
                        help="Use explicitly given branch as upstream")
    parser.add_argument("--filterLists", default=[],
                        nargs='+', help=argparse.SUPPRESS)
    parser.add_argument("--usedefaultfiler", default=True,
                        action="store_false", help=argparse.SUPPRESS)
    RUNARGS = parser.parse_args()
    return RUNARGS



import json
import os
import re

from ensurecvs.configs.keywords_base import KEYWORDS_BASE
from ensurecvs.configs.extension_base import EXTENSION_BASE


class IssueFinder(object):
    def __init__(self, useDefaultFilter=True, filterLists=[]):
        super().__init__()
        self.__filterLists = []
        if useDefaultFilter:
            self.__filterLists += KEYWORDS_BASE
        for fil in filterLists:
            try:
                with open(os.path.join(fil)) as i:
                    self.__filterLists += json.load(i)
            except Exception as e:
                print(str(e))

    def __containsValidFile(self, cvs, commit):
        _cfiles = cvs.GetPatchPathsFromCommit(commit)
        ignorable = True
        for f in _cfiles:
            ignorable &= any([f.endswith(x) for x in EXTENSION_BASE])
        return not ignorable

    def __classifyCommitMessage(self, message):
        __hits = {k: 0 for k in KEYWORDS_BASE.keys()}
        for k, v in KEYWORDS_BASE.items():
            for word in v:
                if re.search(word, message, re.IGNORECASE):
                    __hits[k] += 1
        return [k for k, v in __hits.items() if v == __hits[max(__hits, key=__hits.get)]]

    def GetIssues(self, csv, commit_list):
        res = []
        for commit in commit_list:
            if not self.__containsValidFile(csv, commit):
                continue
            if self.__classifyCommitMessage(commit.message) in [["corrective"], ["corrective", "streamline"], ["corrective", "management"]]:
                res.append(commit)
        return res



import git

from ensurecvs.classes.CSV_base import CVS


class CVSgit(CVS):

    def __init__(self, localdir):
        super().__init__()
        self.__git = git.Repo(path=localdir)
        self.__git.git.fetch()

    def GetLocalDir(self):
        return self.__git.working_dir

    def GetCurrentRevision(self, branch=None):
        _branch = branch or self.GetCurrentBranch()
        x = [x for x in self.__git.heads if x.name == _branch]
        if x:
            return x[0].commit.hexsha
        return None

    def GetRevisionFromTag(self, tag):
        for x in self.__git.tags:
            for y in x.list_items(self.__git):
                if str(y) == tag:
                    return y.commit.hexsha
        return None

    def GetCurrentBranch(self):
        return self.__git.active_branch

    def GetFirstBranchCommit(self, branch=None):
        _branch = branch or self.GetCurrentBranch()
        if _branch == "master":
            return []
        return list(self.__git.iter_commits("master..{}".format(_branch)))[-1]

    def GetCommitSinceBranch(self, branch=None, master="master"):
        _branch = branch or self.GetCurrentBranch()
        if _branch == master:
            return []
        _commit = self.GetFirstBranchCommit(branch=_branch)
        _master_commits = self.GetAllBranchCommits(branch=master)
        _start = None
        for i in _commit.parents:
            if any([True for x in _master_commits if x.binsha == i.binsha]):
                _start = i.hexsha
                break
        if _start is not None:
            return self.GetUpstreamChangeSet(branch=master, start=_start)
        return []

    def GetPatchPathsFromCommit(self, commit):
        res = set()
        for x in self.__git.git.diff_tree("--no-commit-id", "--name-only", "-r", "{}".format(commit.hexsha)).split("\n"):
            res.add(x)
        return list(res)

    def GetCherryPickedCommit(self, commit, start, branch=None):
        _branch = branch or self.GetCurrentBranch()
        _end = self.GetCurrentRevision(_branch)
        for i in self.__git.iter_commits("{}..{}".format(start, _end)):
            if i.message == commit.message:
                return i
            if i.summary == commit.summary and \
                    self.GetPatchPathsFromCommit(i) == self.GetPatchPathsFromCommit(commit):
                return i
        return None

    def GetAllBranchCommits(self, branch=None):
        _branch = branch or self.GetCurrentBranch()
        return self.__git.iter_commits("{}".format(_branch))

    def GetUpstreamChangeSet(self, start=None, end=None, branch=None):
        _end = end or self.GetCurrentRevision(branch=branch)
        _start = start or _end
        return list(self.__git.iter_commits("{}..{}".format(_start, _end)))



KEYWORDS_BASE = {
    "corrective": [
        "BROKE",
        "BUG",
        "CORRECT",
        "DEPRACE",
        "ERROR",
        "FAIL",
        "FIX",
        "ISSUE",
        "PATCH",
        "PROBLEM",
        "PROPER",
        "WRONG",
        "PREVENT",
        "AGAIN",
    ],
    "feature": [
        "ADD",
        "CREATE",
        "FEATURE",
        "INCLUDE",
        "INCREASE",
        "INITIAL",
        "INTRODUCE",
        "NEW",
        "REQUEST",
        "REQUIREMENT",
        "START",
    ],
    "management": [
        "CI",
        "COMMENT",
        "DOC",
        "FORMATTING",
        "LICENSE",
        "MANUAL",
        "MERGE",
        "MIGRAT",
        "ORGANIZ",
        "README",
        "RELEASE",
        "STRUCTURE",
        "TODO",
        "UPGRADE",
        "REVER",
        "TEMPORARY"
    ],
    "streamline": [
        "ADJUST",
        "ASSET",
        "BETTER",
        "CHANG",
        "CLEAN",
        "COVERAGE",
        "DELET",
        "ELIMINAT",
        "ENHANC",
        "IMPROV",
        "MISS",
        "MODIF",
        "OBSOLETE",
        "OPTIMIZ",
        "POLISH",
        "PRETT",
        "REARRANG",
        "REFACTOR",
        "REMOV",
        "RENAM",
        "REPLAC",
        "RESTRUCTUR",
        "SIMPLIFY",
        "TEST",
        "UNIT",
        "UPDATE",
        "VERBOSE",
    ]
}
